'''
本库用于科学计算和快速建模

maysics主要包括六个模块：

1、algorithm 储存了几种模拟方法，用于简易模拟；
2、constant 储存了部分常数；
3、models 储存了几种常用的数学物理定律、方程、模型以便快速构建数理模型；
4、model_selection 用于评估和选择模型；
5、operator 储存了部分常见的算符算子，辅助数学运算；
6、transformation 储存了常用的坐标转换及其他数学变换。


This package is used for calculating and fast modeling.

maysics includes six modules:

1. "algorithm" stores several simulation methods for simple simulation;
2. "constant" contents some usual constants;
3. "models" stores several commonly used laws, equations and models of mathematical physics for fast modeling;
4. "model_selection" used for estimating and selecting model;
5. "calculate" stores some common operators to assist in mathematical operations;
6. "transformation" stores common coordinate transformations and other mathematical transformations.

maysics-|- __init__
        |- algorithm -------|- GA
        |                   |- MC
        |                   |- SA
        |
        |- constant --------|- LP
        |
        |- model_selection -|- Error
        |                   |- Estimate
        |                   |- Search
        |                   |- Sense
        |
        |- models ----------|- ED
        |                   |- Fouriers_law
        |                   |- Leslie
        |                   |- Logistic
        |                   |- MVD_law
        |                   |- Plancks_law
        |
        |- operator --------|- Del
        |                   |- Dif
        |                   |- H
        |                   |- Laplace
        |
        |- transformation --|- Cylinder
        |                   |- Lorentz
        |                   |- Polar
        |                   |- Sphere
'''

import numpy as np


def arr(f):
    '''
    将矢量函数的输出形式统一为ndarray
    
    返回值：更改输出格式后的函数
    
    
    transform the output of vector function as ndarray
    
    return: function after changing the output format
    '''
    def obj(x):
        func = f(x)
        try:
            return np.array(func)
        except:
            return func
    return obj


def add(*arg):
    '''
    实现函数与同型函数、函数与数之间的加法
    要求作用函数若输出列表，必须是ndarray格式
    
    返回值：相加后的新函数
    
    
    addition between function and function or function and number
    if output of the function is list, it requires ndarray
    
    return: new function after addition
    '''
    def obj(x):
        list = []
        for i in range(len(arg)):
            if type(arg[i]).__name__ == 'function':
                list.append(arg[i](x))
            else:
                list.append(arg[i])
        return sum(list)
    return obj


def mul(*arg):
    '''
    实现函数与同型函数、函数与数之间的乘法
    要求作用函数若输出列表，必须是ndarray格式
    
    返回值：相乘后的新函数
    
    
    multiplication between function and function or function and number
    if output of the function is list, it requires ndarray
    
    return: new function after multiplication
    '''
    def obj(x):
        result = 1
        for i in range(len(arg)):
            if type(arg[i]).__name__ == 'function':
                result *= arg[i](x)
            else:
                result *= arg[i]
        return result
    return obj


def sub(minuend, subtrahend):
    '''
    实现函数与同型函数、函数与数之间的减法
    要求作用函数若输出列表，必须是ndarray格式
    minuend：被减数
    subtrahend：减数
    
    返回值：相乘后的新函数
    
    
    subtraction between function and function or function and number
    if output of the function is list, it requires ndarray
    minuend: minuend
    subtrahend: subtrahend
    
    return: new function after subtraction
    '''
    def obj(x):
        if type(minuend).__name__ == 'function':
            result_of_minuend = minuend(x)
        else:
            result_of_minuend = minuend
        if type(subtrahend).__name__ == 'function':
            result_of_subtrahend = subtrahend(x)
        else:
            result_of_subtrahend = subtrahend
        
        result = result_of_minuend - result_of_subtrahend
        return result
    return obj


def divi(dividend, divisor):
    '''
    实现函数与同型函数、函数与数之间的减法
    要求作用函数若输出列表，必须是ndarray格式
    
    参数
    ----
    dividend：被除数
    divisor：除数
    
    返回值：相乘后的新函数
    
    
    Parameters
    ----------
    division between function and function or function and number
    if output of the function is list, it requires ndarray
    dividend: dividend
    divisor: divisor
    
    return: new function after division
    '''
    def obj(x):
        if type(dividend).__name__ == 'function':
            result_of_dividend = dividend(x)
        else:
            result_of_dividend = dividend
        if type(divisor).__name__ == 'function':
            result_of_divisor = divisor(x)
        else:
            result_of_divisor = divisor
        result = result_of_dividend - result_of_divisor
        return result
    return obj


def shuffle(*arg):
    '''
    以相同方法打乱多个序列或打乱一个序列
    
    返回值：一个ndarray
    
    
    shuffle multiple sequences in the same way or shuffle a sequences
    
    return: a ndarray
    '''
    state = np.random.get_state()
    a_new_list = []
    for li in arg:
        np.random.set_state(state)
        np.random.shuffle(li)
        a_new_list.append(li)
    return np.array(a_new_list)


def r(*arg):
    '''
    相关系数
    
    返回值：各数组之间的相关系数矩阵
    
    
    correlation coefficient
    
    return: matrix of correlation coefficient
    '''
    arg = np.array(arg, dtype=float)
    if len(arg.shape) != 2:
        raise Exception("Input list should be 1 dimension.")
    
    cov_mat = np.cov(arg)
    var_mat = np.diagonal(cov_mat)**0.5
    
    for i in range(cov_mat.shape[0]):
        cov_mat[i] /= var_mat[i]
        cov_mat[:, i] /= var_mat[i]
    
    return cov_mat


def data_split(data, targets, train_size=None, test_size=None, shuffle=True, random_state=None):
    '''
    分离数据
    
    参数
    ----
    data：数据
    targets：指标
    train_size：浮点数类型，可选，训练集占总数据量的比，取值范围为(0, 1]，默认为0.75
    test_size：浮点数类型，可选，测试集占总数据量的比，取值范围为[0, 1)，当train_size被定义时，该参数无效
    shuffle：布尔类型，可选，True表示打乱数据，False表示不打乱数据，默认为True
    random_state：整型，可选，随机种子
    
    
    split the data
    
    Parameters
    ----------
    data: data
    targets: targets
    train_size: float, callable, ratio of training set to total data, value range is (0, 1], default=0.75
    test_size: float, callable, ratio of test set to total data, value range is [0, 1)
    shuffle: bool, callable, 'True' will shuffle the data, 'False' will not, default = True
    random_state: int, callable, random seed
    '''
    if not (train_size or test_size):
        train_size = 0.75
    elif test_size:
        train_size = 1 - test_size
    
    if train_size <= 0 or train_size > 1:
        raise Exception("'train_size' should be in (0, 1], 'test_size' should be in [0, 1)")
    
    if shuffle:
        np.random.seed(random_state)
        
        state = np.random.get_state()
        np.random.shuffle(data)
        np.random.set_state(state)
        np.random.shuffle(targets)
        
    num_of_data = len(data)
    train_data = data[:int(num_of_data * train_size)]
    train_target = targets[:int(num_of_data * train_size)]
    validation_data = data[int(num_of_data * train_size):]
    validation_target = targets[int(num_of_data * train_size):]
    
    return train_data, train_target, validation_data, validation_target


def kfold(data, targets, n, k=5):
    '''
    参数
    ----
    data：数据
    targets：指标
    n：整型，表示将第n折作为验证集，从0开始
    k：整型，可选，k折验证的折叠数，默认k=5
    
    返回值：训练集和验证集的元组
    
    
    Parameters
    ----------
    data: data
    targets: targets
    n: int, take the nth part as validation set, starting from 0
    k: int, callable, the number of k-fold, default = 5
    
    return: the tuple of training set and validation set
    '''
    num_validation_samples = len(data) // k
    
    validation_data = data[num_validation_samples * n:
                           num_validation_samples * (n + 1)]
    validation_targets = targets[num_validation_samples * n:
                                 num_validation_samples * (n + 1)]
    
    train_data = np.concatenate([data[: num_validation_samples * n],
                                 data[num_validation_samples * (n + 1):]])
    train_targets = np.concatenate([targets[: num_validation_samples * n],
                                    targets[num_validation_samples * (n + 1):]])
    
    return train_data, train_targets, validation_data, validation_targets


'''
本模块储存着一些常用的常数

This module contents some usual constants

数学方面 / math

chaos_1: 第一费根鲍姆常数 / the first Feigenbaum constant
chaos_2: 第二费根鲍姆常数 / the second Feigenbaum constant
e: 自然常数 / natural constant
gamma: 欧拉-马歇罗尼常数 / Euler-Mascheroni constant
golden: 黄金比例 / golden ratio
K: 兰道-拉马努金常数 / Landau–Ramanujan constant
K0: 辛钦常数 / Khintchine constant
pi: 圆周率 / Ratio of circumference to diameter
LP(): 勒让德多项式 / Legendre Polynomials


物理方面 / physics

AU: 天文单位 / astronomical unit
atm: 标准大气压 / standard atmospheric pressure
c: 真空光速 / the speed of light in vacuum
c_e: 元电荷 / elementary charge
epsilon: 真空介电常数 / permittivity of vacuum
g: 重力加速度 / gravitational acceleration
G: 万有引力常数 / gravitational constant
h: 普朗克常数 / Planck constant
hr: 约化普朗克常数 / reduced Planck constant
k: 玻尔兹曼常数 / Boltzmann constant
lambdac: 康普顿波长 / Compton wavelength
ly: 光年 / light-year
m_e: 电子质量 / mass of electron
m_earth: 地球质量 / mass of the earth
m_n: 中子质量 / mass of neutron
m_p: 质子质量 / mass of proton
m_s: 太阳质量 / mass of the sun
miu: 真空磁导率 / permeability of vacuum
NA: 阿伏伽德罗常数 / Avogadro constant
pc: 秒差距 / Parsec
Platonic_year: 柏拉图年 / Platonic year
R: 理想气体常数 / gas constant
r_earth: 地球平均半径 / mean radius of the earth
r_sun: 太阳平均半径 / mean radius of the sun
r_e_m: 地月平均距离 / mean distance between the earth and the moon
SB: 斯特藩-玻尔兹曼常量 / Stefan-Boltzmann constant
v1: 第一宇宙速度 / first cosmic velocity
v2: 第二宇宙速度 / second cosmic velocity
v3: 第三宇宙速度 / third cosmic velocity
'''

import numpy as np
from scipy.special import factorial


# math
pi = 3.141592653589793
e = 2.718281828459045
golden = 1.618033988749895            #黄金比例
gamma = 0.57721566490153286060651209  #欧拉-马歇罗尼常数
K = 0.76422365358922066299069873125   #兰道-拉马努金常数
chaos_1 = 4.669201609102990           #第一费根鲍姆常数
chaos_2 = 2.502907875095892           #第二费根鲍姆常数
K0 = 2.6854520010                     #辛钦常数

class LP():
    '''
    勒让德多项式
    
    
    Legendre Polynomials
    '''
    @classmethod
    def N(self, l):
        '''
        l阶勒让德多项式的模
        
        参数
        ----
        l：整型，阶数
        
        返回值：一个数
        
        
        module of Legendre Polynomials at degree l
        
        Paramater
        ---------
        l: int, degree
        
        return: a number
        '''
        return (2 / (2 * l + 1))**0.5
    
    @classmethod
    def P(self, x, l):
        '''
        l阶勒让德多项式的值
        
        参数
        ----
        x：输入值
        l：整型，阶数
        
        返回值：一个数
        
        
        value of Legendre Polynomials at degree l
        
        Parameters
        ----------
        x: input value
        l: int, degree
        
        return: a number
        '''
        result = 0
        for k in range(int(l/2) + 1):
            result += (-1)**k * factorial(2 * l - 2 * k)\
            / 2**l / factorial(k) / factorial(l - k) / factorial(l - 2 * k)\
            * x**(l - 2 * k)
        return result
    
    @classmethod
    def P_mat(self, x, l):
        '''
        l阶勒让德多项式的值的矩阵
        
        参数
        ----
        x: 输入值
        l：列表形式，阶数的起止列表
        
        返回值：numpy.matrix
        
        
        matrix of values of Legendre Polynomials at degree l
        
        Parameters
        ----------
        x: input value
        l: list, the begin l and the end l
        
        return: numpy.matrix
        '''
        result_matrix = []
        for i in range(l[0], l[1] + 1):
            result = 0
            for k in range(int(i/2) + 1):
                result += (-1)**k * factorial(2 * i - 2 * k)\
                / 2**i / factorial(k) / factorial(i - k) / factorial(i - 2 * k)\
                * x**(i - 2 * k)
            result_matrix.append(result)
        return np.mat(result_matrix)
        


# physics
atm = 101325                          #标准大气压
c = 299792458.0
c_e = 1.602176634e-19
m_e = 9.1093837015e-31                #电子质量
m_earth = 5.965e24                    #地球质量
m_n = 1.67492749804e-27               #中子质量
m_p = 1.67262192369e-27               #质子质量
m_sun = 1.9891e30                     #太阳质量
G = 6.6743e-11
g = 9.80665
h = 6.62607015e-34
hr = 1.0545718176461565e-34           #约化普朗克常数
k = 1.380649e-23                      #玻尔兹曼常量
miu = 1.2566370614359173e-06          #真空磁导率
epsilon = 8.85e-12                    #真空介电常数
NA = 6.02214179e23                    #阿伏伽德罗常数
R = 8.31446114959365
r_earth = 6371393                     #地球平均半径
r_sun = 6.955e8                       #太阳平均半径
r_moon = 384403000                    #地月平均距离
lambdac = 2.4263102175e-12            #康普顿波长
ly = 9460730472580800                 #光年
AU = 149597870700                     #天文单位
pc = 3.0856775814671915808e16         #秒差距
Platonic_year = 26000                 #柏拉图年
SB = 5.670367e-8                      #斯特藩-玻尔兹曼常量
v1 = 7900                             #第一宇宙速度
v2 = 11200                            #第二宇宙速度
v3 = 16700                            #第三宇宙速度


'''
本模块储存着部分常用的数理物理方程、定律、数学模型等

This module stores some commonly used mathematical physical equations, laws, mathematical models, etc
'''

import numpy as np
from maysics import operator
from maysics import constant
from scipy.integrate import solve_ivp
from scipy.sparse import csr_matrix


class Fouriers_law(operator.Del):
    '''
    傅里叶定律/热传导定律
    热流量：J=-k▽T
    
    参数
    ----
    k：热导率
    acc：浮点数类型，可选，求导精度，默认为0.1
    
    
    Fourier's Law: J=-k▽T
    
    Parameters
    ----------
    k: coefficient of thermal conductivity
    acc: float, callable, accuracy of derivation, default=0.1
    '''
    def __init__(self, k, acc=0.1):
        self.acc = acc
        self.k = k
    
    
    def fit(self, T):
        '''
        参数
        ----
        T：温度分布函数
        
        
        Parameter
        ---------
        T: temperature distribution function
        '''
        j = operator.Del.grad(self, T)
        def j2(x):
            j1 = -self.k * j(x)
            return j1
        
        self.J = j2



class MVD_law():
    '''
    麦克斯韦速率分布律
    
    
    Maxwell's Velocity Distribution Law
    '''
    @classmethod
    def fit(self, m, T):
        '''
        参数
        ----
        m：气体分子质量, 单位：kg
        T：气体温度, 单位：K
        
        属性
        ----
        fv：速率分布函数
        v_mean：平均速率
        v_p：最概然速率
        v_rms：均方根速率
        
        
        Parameters
        ----------
        m: mass of gas molecule, unit: kg
        T: temperature of gas, unit: K
        
        Attributes
        ----------
        fv: velocity distribution function
        v_mean: average velocity
        v_p: most probable velocity
        v_rms: root-mean-square velocity
        '''
        def func_of_v(v):
            f_v_1 = 4 * np.pi * v**2
            f_v_2 = m / (2 * np.pi * constant.k * T)**1.5
            f_v_3 = np.e**(-m * v**2 / (2 * constant.k * T))
            return f_v_1 * f_v_2 * f_v_3
        
        self.fv = func_of_v
        v_part = (constant.k * T / (np.pi * m))**0.5
        self.v_mean = 8**0.5 * v_part
        self.v_p = 2**0.5 * v_part
        self.v_rms = 3**0.5 * v_part



class Plancks_law():
    '''
    普朗克黑体辐射定律
    
    属性
    ----
    Mf：频率形式的普朗克公式，频率单位：10^10 kHz
    Ml：波长形式的普朗克公式，波长单位：100nm
    
    
    Planck's Blackbody Radiation Law
    
    Attributes
    ----------
    Mf: Planck's formula in the form of frequency, unit of frequency: 10^10 kHz
    Ml: Planck's formula in the form of wave length, unit of wave length: 100 nm
    '''
    @classmethod
    def fit(self, T):
        '''
        参数
        ----
        T：黑体温度，单位：K
        
        
        Parameter
        ---------
        T: temperature of blackbody, unit: K
        '''
        h_k = constant.h / constant.k
        def Mf(f):
            f = f * 1e13
            f_1 = 2 * np.pi * constant.h * f**3 / constant.c**2
            f_2 = 1 / (np.e**(h_k * f / T) - 1)
            return f_1 * f_2
        
        def Ml(l):
            l = l * 1e-7
            l_1 = 2 * np.pi * constant.h * constant.c**2 / l**5
            l_2 = 1 / (np.e**(h_k * constant.c / (l * T)) - 1)
            return l_1 * l_2
        
        self.Mf = Mf
        self.Ml = Ml



class ED():
    '''
    传染病模型
    本类可构建SI模型、SIR模型、SIRS模型、SEIR模型
    S是易感染者
    E是暴露者，染病但还不具有传染能力
    I是感染者，带有传染能力
    R是康复者(带有免疫力)或死者
    
    参数
    ----
    I0：初始感染者
    K：样本总数
    beta：感染率
    
    
    mathematical models of epidemic diseases
    this class is used to buil SI, SIR, SIRS, SEIR
    S means susceptible people
    E means exposed people, which are infected without ability to infect
    I means infected people with ability to infect
    R means recovered people with immunity or dead people
    
    Parameters
    ----------
    I0: initial infected people
    K: sample size
    beta: infection rate
    '''
    def __init__(self, I0, K, beta, R=0, E=0):
        self.I0 = I0
        self.K = K
        self.beta = beta
        self.R = R
        self.E = E
        self.S = K - I0 - R - E
    
    
    def SI(self):
        '''
        SI模型：
        该模型不需要再额外输入参数，且可得到解析解
        解的形式为：I = K*(1-(K-I0)/(K+I0+I0*np.e**(beta*K*t)))

        返回值：元组形式，(I(t), S(t))
        
        
        SI:
        no more parameters
        solution: I = K*(1-(K-I0)/(K+I0+I0*np.e**(beta*K*t)))
        
        return: tuple, (I(t), S(t))
        '''
        def obj(t):
            I_pre_down_ = self.K + self.I0 * (np.e**(self.beta * self.K * t) - 1)
            I_pre_up_ = self.K - self.I0
            I_func_ = self.K * (1 - I_pre_up_ / I_pre_down_)
            S_func_ = self.K - I_func_
            return I_func_, S_func_
        return obj
    
    
    def SIR(self, gamma, t_span, method='RK45' ,t_eval=None):
        '''
        SIR模型：
        该模型不能得到解析解，给出的是基于solve_ivp得到的数值解数组
        
        参数
        ----
        gamma：康复率
        t_span：元组形式，求解的上下限
        method：字符串形式，可选，求解方法，可选择'RK45'、'RK23'、'DOP835'、'Radau'、'BDF'、'LSODA'
        t_eval：数组形式，可选，每当t等于该数组中的值时，会生成一个数值解

        返回值：solve_ivp数值解，顺序是I, S, R
        
        
        SIR:
        
        Parameters
        ----------
        gamma: recovery rate
        t_span: tuple, limits of solution
        method: str, callable, solving method, 'RK45', 'RK23', 'DOP835', 'Radau', 'BDF', 'LSODA' are optional
        t_eval: list, callable, when 't' equals the value in the list, it will generate a numerical solution
        
        return: numerical solution of solve_ivp, the order is I, S, R
        '''
        y0 = [self.I0, self.S, self.R]
        def epis_equas(t, x):
            y1 = self.beta * x[0] * x[1] - gamma * x[0]
            y2 = -self.beta * x[0] * x[1]
            y3 = gamma * x[0]
            return y1, y2, y3
        result = solve_ivp(epis_equas, t_span=t_span, y0=y0, method=method, t_eval=t_eval)
        return result
    
    
    def SIRS(self, gamma, alpha, t_span, method='RK45' ,t_eval=None):
        '''
        SIRS模型：
        该模型不能得到解析解，给出的是基于solve_ivp得到的数值解数组
        
        参数
        ----
        gamma：康复率
        alpha：衡量康复者获得免疫的时间
        t_span：元组形式，求解的上下限
        method：字符串形式，可选，求解方法，可选择'RK45'、'RK23'、'DOP835'、'Radau'、'BDF'、'LSODA'
        t_eval：数组形式，可选，每当t等于该数组中的值时，会生成一个数值解

        返回值：solve_ivp数值解，顺序是I, S, R
        
        
        SIRS:
        
        Parameters
        ----------
        gamma: recovery rate
        alpha: measuring the time that the recovered people with immunity
        t_span: tuple, limits of solution
        method: str, callable, solving method, 'RK45', 'RK23', 'DOP835', 'Radau', 'BDF', 'LSODA' are optional
        t_eval: list, callable, when 't' equals the value in the list, it will generate a numerical solution
        
        return: numerical solution of solve_ivp, the order is I, S, R
        '''
        y0= [self.I0, self.S, self.R]
        def epis_equas(t, x):
            y1 = self.beta * x[0] * x[1] - gamma * x[0]
            y2 = -self.beta * x[0] * x[1] + alpha * x[2]
            y3 = gamma * x[0] - alpha * x[2]
            return y1, y2, y3
        result = solve_ivp(epis_equas, t_span=t_span, y0=y0, method=method, t_eval=t_eval)
        return result
    
    
    def SEIR(self, gamma1, gamma2, alpha, t_span, method='RK45' ,t_eval=None):
        '''
        SEIR模型：
        该模型不能得到解析解，给出的是基于solve_ivp得到的数值解数组
        
        参数
        ----
        gamma1：潜伏期康复率
        gamma2：患者康复率
        alpha：衡量康复者获得免疫的时间
        t_span：元组形式，求解的上下限
        method：字符串形式，可选，求解方法，可选择'RK45'、'RK23'、'DOP835'、'Radau'、'BDF'、'LSODA'
        t_eval：数组形式，可选，每当t等于该数组中的值时，会生成一个数值解

        返回值：solve_ivp数值解，顺序是I, S, R, E
        
        
        SEIR:
        
        Parameters
        ----------
        gamma1: recovery rate of incubation
        gamma2: recovery rate of patients
        alpha: measuring the time that the recovered people with immunity
        t_span: tuple, limits of solution
        method: str, callable, solving method, 'RK45', 'RK23', 'DOP835', 'Radau', 'BDF', 'LSODA' are optional
        t_eval: list, callable, when 't' equals the value in the list, it will generate a numerical solution
        
        return: numerical solution of solve_ivp, the order is I, S, R, E
        '''
        y0 = [self.I0, self.S, self.R, self.E]
        def epis_equas(t, x):
            y1 = alpha * x[3] - gamma2 * x[0]
            y2 = -self.beta * x[0] *x[1]
            y3 = gamma1 * x[3] + gamma2 * x[0]
            y4 = self.beta * x[0] * x[1] - (alpha + gamma1) * x[3]
            return y1, y2, y3, y4
        result = solve_ivp(epis_equas, t_span=t_span, y0=y0, method=method, t_eval=t_eval)
        return result



class Leslie():
    '''
    Leslie模型
    解的形式为：Nt = M**t * N0
    其中，Nt是t时刻的个体数列表，M是Leslie矩阵
    
    参数
    ----
    N0：列表，各年龄层初始个体数目
    r：数，各年龄层的生殖率
    s：数，各年龄层到下一个年龄层的存活率
    age_range：整型，年龄段的跨度，默认为1
    
    
    Leslie model
    solution: Nt = M**t * N0
    Nt is the list of individuals at time 't' , M is Leslie matrix
    
    Parameters
    ----------
    N0: list, initial number of individuals in each age group
    r: number, reproductive rate of each age group
    s: number, survival rate to next age group of each group
    age_range: int, the span of age groups, default=1
    '''
    def __init__(self, N0, r, s, age_range=1):
        self.N0 = np.mat(N0).T
        self.r = np.array(r)
        self.s = np.array(s)
        self.age_range = age_range
        
        Leslie_matrix = np.zeros((len(s), len(N0)))
        Leslie_matrix = np.vstack((r, Leslie_matrix))
        for i in range(1, len(N0)):
            Leslie_matrix[i][i-1] = s[i-1]
        self.Leslie_matrix = csr_matrix(Leslie_matrix)
    
    
    def predict(self, t):
        '''
        参数
        ----
        t：时间
        
        返回值：Nt
        
        
        Parameter
        ---------
        t: time
        
        return: Nt
        '''
        t_times = t // self.age_range
        dense_Leslie_matrix = self.Leslie_matrix.todense()
        Nt = dense_Leslie_matrix**t_times * self.N0
        return Nt



class Logistic():
    '''
    Logistic人口增长模型
    该模型可得到解析解
    解的形式为：Nt = K*N0/(N0+(K-N0)*np.e**(-r*t))
    其中，Nt是t时刻的人口数
    
    参数
    ----
    N0：数，现有人口数
    r：数，人口自然增长率
    K：数，环境资源允许的稳定人口数
    
    
    Logisyic population growth models
    solution: Nt = K*N0/(N0+(K-N0)*np.e**(-r*t))
    Nt is the population at time 't'
    
    Parameters
    ----------
    N0: number or list, initial population
    r: number, natural population growth rate
    K: number, stable population allowed by environmental resources
    '''
    def __init__(self, N0, r, K):
        self.N0 = N0
        self.r = r
        self.K = K
    
    
    def predict(self, t):
        '''
        参数
        ----
        t：时间
        
        返回值：Nt
        
        
        Parameter
        ---------
        t: time
        
        return: Nt
        '''
        Nt_pre_down_ = self.N0 + (self.K - self.N0) * np.e**(-self.r * t)
        Nt = self.K * self.N0 / Nt_pre_down_
        return Nt


'''
本模块储存着几种常用坐标变换

This module stores several common coordinate transformations
'''

import numpy as np
from maysics import constant


class Lorentz():
    '''
    洛伦兹变换
    
    参数
    ----
    v：速度
    
    
    Lorentz transformation
    
    Parameters
    ----------
    v: speed
    '''
    def __init__ (self, v):
        self.v = v
        self.γ = 1/(1-(v/constant.c)**2)**0.5
    

    def lor(self, x):
        '''
        洛伦兹坐标正变换
        
        参数
        ----
        x：列表，(x, y, z, t)
        
        coordinate transformation, S → S'
        
        Parameters
        ----------
        x: list, (x, y, z, t)
        '''
        xp = self.γ * (x[0] - self.v * x[3])
        yp = x[1]
        zp = x[2]
        tp = self.γ * (x[3] - self.v * x[0] / constant.c**2)
        result = np.array([xp, yp, zp, tp])
        return result
    

    def ilor(self, xp):
        '''
        洛伦兹坐标逆变换
        
        参数
        ----
        xp：列表，(x', y', z', t')
        
        coordinate transformation, S' → S:
        
        Parameters
        ----------
        xp: list, (x', y', z', t')
        '''
        x = self.γ * (xp[0] + self.v * xp[3])
        y = xp[1]
        z = xp[2]
        t = self.γ * (xp[3] + self.v * xp[0] / constant.c**2)
        result = np.array([x, y, z, t])
        return result
    

    def vel(self, vo):
        '''
        洛伦兹速度正变换
        
        参数
        ----
        vo：列表，(vx, vy, vz)
        
        speed transformation, S → S'
        
        Parameters
        ----------
        v0: list, (vx, vy, vz)
        '''
        factor = 1 - self.v * vo[0] / constant.c**2
        vxp = (vo[0] - self.v) / factor
        vyp = vo[1] * self.γ / factor
        vzp = vo[2] * self.γ / factor
        result = np.array([vxp, vyp, vzp])
        return result
    

    def ivel(self, vp):
        '''
        洛伦兹速度逆变换
        
        参数
        ----
        vp：列表，(vx', vy', vz')
        
        speed transformation, S' → S
        
        Parameters
        ----------
        vp: list, (vx', vy', vz')
        '''
        factor = 1 + self.v * vp[0] / constant.c**2
        vxo = (vp[0] + self.v) / factor
        vyo = vp[1] * self.γ / factor
        vzo = vp[2] * self.γ / factor
        result = np.array([vxo, vyo, vzo])
        return result


class Polar():
    '''
    极坐标系与直角坐标系转换
    
    
    polar transformation
    '''
    @classmethod
    def pol(self, x):
        '''
        极坐标正变换
        
        参数
        ----
        x：列表，(x, y)
        
        
        Parameters
        ----------
        x: list, (x, y)
        '''
        r = (x[0]**2 + x[1]**2)**0.5
        if x[0] == 0:
            if x[1] > 0:
                theta = np.pi/2
            elif x[1] < 0:
                theta = -np.pi/2
            elif x[1] == 0:
                theta = 0
        else:
            theta = np.arctan(x[1] / x[0])
        result = np.array([r, theta])
        return result


    @classmethod
    def ipol(self, x):
        '''
        极坐标逆变换
        
        参数
        ----
        x：列表，(r, θ)
        
        
        Parameters
        ----------
        x: list, (r, θ)
        '''
        x = x[0] * np.cos(x[1])
        y = x[0] * np.sin(x[1])
        result = np.array([x, y])
        return result


class Cylinder():
    '''
    柱坐标系与直角坐标系转换
    
    cylinder transformation
    '''
    @classmethod
    def cyl(self, x):
        '''
        柱坐标正变换
        
        参数
        ----
        x：列表，(x, y, z)
        
        
        Parameters
        ----------
        x: list, (x, y, z)
        '''
        r = (x[0]**2 + x[1]**2)**0.5
        if x[0] == 0:
            if x[1] > 0:
                theta = np.pi/2
            elif x[1] < 0:
                theta = -np.pi/2
            elif x[1] == 0:
                theta = 0
        else:
            theta = np.arctan(x[1] / x[0])
        z = x[2]
        result = np.array([r, theta, z])
        return result


    @classmethod
    def icyl(self, x):
        '''
        柱坐标逆变换
        
        参数
        ----
        x：列表，(r, θ, z)
        
        
        Parameters
        ----------
        x: list, (r, θ, z)
        '''
        x = x[0] * np.cos(x[1])
        y = x[0] * np.sin(x[1])
        z = x[2]
        result = np.array([x, y, z])
        return result


class Sphere():
    '''
    球坐标系与直角坐标系转换
    
    sphere transformation
    '''
    @classmethod
    def sph(self, x):
        '''
        球坐标正变换
        
        参数
        ----
        x：列表，(x, y, z)
        
        
        Parameters
        ----------
        x: list, (x, y, z)
        '''
        r = (x[0]**2 + x[1]**2 + x[2]**2)**0.5
        if r == 0:
            theta = 0
            phai = 0
        else:
            p = np.arccos(x[2] / r)
            if x[0] == 0:
                if x[1] > 0:
                    theta = np.pi/2
                elif x[1] < 0:
                    theta = -np.pi/2
                elif x[1] == 0:
                    theta = 0
            else:
                theta = np.arctan(x[1] / x[0])
        result = np.array([r, theta, phai])
        return result


    @classmethod
    def isph(self, x):
        '''
        球坐标逆变换
        
        参数
        ----
        x：列表，(r, θ, φ)
        
        
        Parameters
        ----------
        x: list, (r, θ, φ)
        '''
        x = x[0] * np.sin(x[2]) * np.cos(x[1])
        y = x[0] * np.sin(x[2]) * np.sin(x[1])
        z = x[0] * np.cos(x[2])
        result = np.array([x, y, z])
        return result


'''
本模块储存了蒙特卡洛模拟、遗传算法和模拟退火算法，用于简易模拟

This module stores Monte Carlo, Genetic Algorithm and Simulated Annealing Algorithm for simple simulation
'''

import random
import numpy as np


class MC():
    '''
    蒙特卡洛模拟
    
    参数
    ----
    loop：整型，可选，循环次数，默认为1000
    random_type：字符串类型，可选，可以取'random'和'randint'，默认'random'
    random_state：整型，可选，随机种子
    begin：整型，可选，表示整数随机序列的开始数字，仅当random_type='randint'时起作用
    end：整型，可选，表示整数随机序列的结束数字，仅当random_type='randint'时起作用
    
    属性
    ----
    loop：循环次数
    random_type：取随机的方法
    EX：select函数输出值的数学期望
    DX：select函数输出值的方差
    
    
    Monte Carlo
    
    Parameters
    ----------
    loop: int, callable, loop count, default = 10000
    random_type: str, callable, 'random' and 'randint' are optional, default = 'random'
    random_state: int, callable, random seed
    begin: int, callable, beginning number of the random sequence, it's used only when random_type='randint'
    end: int, callable, end number of the random sequence, it's used only when random_type='randint'
    
    Atrributes
    ----------
    loop: loop count
    random_type: the method of generating random number
    EX: mathematical expectation of output of select function
    DX: variance of output of select function
    '''
    def __init__(self, loop=1000, random_type='random', random_state=None, begin=None, end=None):
        self.loop = loop
        self.random_type = random_type
        self.__begin = begin
        self.__end = end
        np.random.seed(random_state)
    

    def sel_ind(self, condition):
        '''
        用于快速构建条件函数
        将condition函数作用于随机序列全部元素
        任意一个元素满足condition，则会输出1，否则输出0
        
        参数
        ----
        condition：函数，判断一个元素是否符合条件的函数，要求符合则输出1(True)，不符合则输出0(False)

        返回值：一个新函数


        for quick construction of condition function
        apply condition function to all elements of random sequence
        if any element is qualified, 1(True) will be reutrn, otherwise 0(False) will be return
        
        Parameters
        ----------
        condition: function, a function that determines if an element is qualified. If qualified, return 1 (true), otherwise reutrn 0 (false)

        return: a new function
        '''
        def obj(x):
            for i in x:
                if condition(i):
                    return 1
            return 0
        return obj

    
    def sel_any(self, n, condition):
        '''
        用于快速构建条件函数
        将condition函数作用于随机序列全部元素
        至少任意n个元素满足condition，则会输出1，否则输出0
        
        参数
        ----
        n：int类型
        condition：函数，判断一个元素是否符合条件的函数，要求符合则输出1(True)，不符合则输出0(False)

        返回值：一个新函数


        for quick construction of condition function
        apply condition function to all elements of random sequence
        if at least any_n elements are qualified, 1(True) will be reutrn, otherwise 0(False) will be return
        
        Parameters
        ----------
        n: int
        condition: function, a function that determines if an element is qualified. If qualified, return 1 (true), otherwise reutrn 0 (false)

        return: a new function
        '''
        def obj(x):
            calcu_ = 0
            for i in x:
                if condition(i):
                    calcu_ += 1
                if calcu_ >= n:
                    break
            if calcu_ >= n:
                return 1
            else:
                return 0
        return obj
    
    
    def sel_con(self, n, condition):
        '''
        用于快速构建条件函数
        将condition函数作用于随机序列全部元素
        至少连续n个元素满足condition，则会输出1，否则输出0
        
        参数
        ----
        n：int类型
        condition：函数，判断一个元素是否符合条件的函数，要求符合则输出1(True)，不符合则输出0(False)

        返回值：一个新函数


        for quick construction of condition function
        apply condition function to all elements of random sequence
        if at least con_n consecutive elements are qualified, 1(True) will be reutrn, otherwise 0(False) will be return
        
        Parameters
        ----------
        n: int
        condtition: function, a function that determines if an element is qualified. If qualified, return 1 (true), otherwise reutrn 0 (false)
        
        return: a new function
        '''
        def obj(x):
            calcu_ = 0
            for i in x:
                if condition(i):
                    calcu_ += 1
                else:
                    calcu_ *= 0
                if calcu_ >= n:
                    break
            if calcu_ >= n:
                return 1
            else:
                return 0
        return obj
    
    
    def __one_experiment(self, length, dimen, select):
        '''
        进行一次完整的实验
        '''
        if self.random_type == 'random':
            random_list_ = np.random.rand(length, dimen)

        elif self.random_type == 'randint':
            random_list_ = np.random.randint(self.__begin, self.__end, size=(length, dimen))
        
        else:
            raise Exception("'random_type' must be one of 'random' and 'randint'.")

        #如果随机序列满足select，就算实验成功输出1，否则算失败输出0
        if type(select).__name__ == 'function':
            return select(random_list_)
        else:
            judge_ = 1
            for li in select:
                if li(random_list_):
                    judge_ *= 1
                else:
                    judge_ *= 0
            return judge_


    def fit(self, length, dimen, select):
        '''
        进行蒙特卡洛模拟
        
        参数
        ----
        length：整型，随机序列长度
        dimen：整型，随机序列元素维度
        select：条件函数或函数列表，select函数应该以length * dim的二维ndarray为自变量
            select函数用于甄别某一个随机矩阵是否符合预期，符合输出1，不符合输出0
            select函数同时也可以输出其他值以满足实际需求
            select若为函数列表，则要求每个函数只能输出1或0

        
        simulate
        
        Parameters
        ----------
        length: int, the length of random sequence
        dimen: int, dimension of elements in random sequence
        select: function or function list, select function should take 2-dimension ndarray( length * dim) as independent variable
            select function used for Identifying whether the random matrix meets the expectation, if meets, output 1, otherwise output 0
            select function can also output other values to meet the actual demand
            if select is a list, every function in the list can only output 1 or 0
        '''
        final_propability_ = []
        for i in range(self.loop):
            freq_ = MC.__one_experiment(self, length=length,\
                dimen=dimen, select=select)
            final_propability_.append(freq_)
        
        self.EX = np.mean(final_propability_)
        self.DX = np.var(final_propability_)


class GA():
    '''
    遗传算法
    
    参数
    ----
    population：整型，种群数
    iteration: 整型，迭代次数（自然选择次数）
    random_type：字符串类型，可选，可以取'random'和'randint'，默认'random'
    select：字符串类型或函数类型，可选，选择个体的方法，可以取'rw'、'st'或者自定义函数，默认'rw'
        'rw'：基于随机接受的轮盘赌选择
        'st'：随机竞争选择
    crossover：字符串类型或函数类型，可选，交叉互换的方法，可以取'uniform'、'point'或者自定义函数，默认'uniform'
        'uniform'：均匀交叉
        'point'：单点及多点交叉
    begin：整型，可选，表示整数随机序列的开始数字，仅当random_type='randint'时起作用
    end：整型，可选，表示整数随机序列的结束数字，仅当random_type='randint'时起作用
    random_state：整型，可选，随机种子
    select_rate：浮点数类型，可选，选择率（存活率），默认0.3
    mutate_rate：浮点数类型，可选，变异率，默认0.05
    crossover_rate：浮点数类型，可选，基因交叉概率，默认0.5
    
    属性
    ----
    population：种群数
    iteration：迭代次数（自然选择次数）
    random_type：取随机的方法
    select：选择个体的方法
    crossover：交叉互换的方法
    dom：优势种群
    dom_fitness：优势种群的适应度
    

    
    Genetic Algorithm
    
    Parameters
    ----------
    population: int, the population size
    iteration: int, the times of iterations( the times of natural selection)
    random_type: str, callable, 'random' and 'randint' are optional, default = 'random'
    select: string or function, callable, the method of selecting individuals, 'rw', 'st' or custom function, default = 'rw'
        'rw': Roulette Wheel Selection
        'st': Stochastic Tournament Selection
    crossover: string or function, callable, the method of crossing over, 'uniform', 'point' or custom function, default = 'uniform'
        'uniform': Uniform Crossover
        'point': Multi-point Crossover
    begin: int, callable, beginning number of the random sequence, it's used only when random_type='randint'
    end: int, callable, end number of the random sequence, it's used only when random_type='randint'
    random_state: int, callable, random seed
    select_rate: float, callable, selection rate( survival rate), default=0.3
    mutate_rate: float, callable, variation rate, default=0.05
    crossover_rate: float, callable, gene crossover probability, default=0.5
    
    Atrributes
    ----------
    population: the population size
    iteration: the times of iterations( the times of natural selection)
    random_type: the method of generating random number
    select: the method of selecting individuals
    crossover: the method of crossing over
    dom: the dominance
    dom_fitness: the fitness of the dominance
    '''
    def __init__(self, population=1000, iteration=100, random_type='random', select='rw', crossover='uniform', begin=None, end=None, random_state=None, select_rate=0.3, mutate_rate=0.05, crossover_rate=0.5):
        self.population = population
        self.iteration = iteration
        self.random_type = random_type
        self.select = select
        self.crossover = crossover
        self.__begin = begin
        self.__end = end
        self.__select_rate = select_rate
        self.__mutate_rate = mutate_rate
        self.__crossover_rate = crossover_rate
        np.random.seed(random_state)
    
    
    def __mutate_func(self, length, populations_matrix):
        '''
        完成对新一代的变异
        '''
        random_mutate_matrix = np.random.rand(self.population, length)
        for i in range(self.population):
            for j in range(length):
                if random_mutate_matrix <= self.__mutate_rate:
                    if self.random_type == 'random':
                        populations_matrix[i][j] = random.random()
                    elif self.random_type == 'randint':
                        populations_matrix[i][j] = random.randint(self.__begin, self.__end)
    
    
    def __point_crossover(self, num_point, length, parent_matrix):
        '''
        多点交叉
        
        
        Multi-point Crossover
        '''
        if num_point > length:
            raise Exception("'num_point' should be less than 'length'.")
        
        num_of_parents = len(parent_matrix)
        child_matrix = []
        child_population = self.population - num_of_parents
        
        while len(child_matrix) < child_population:
            while True:
                random_num_1 = random.randint(0, num_of_parents)
                random_num_2 = random.randint(0, num_of_parents)
                if random_num_1 != random_num_2:
                    break
            child_individual_1 = parent_matrix[random_num_1]
            child_individual_2 = parent_matrix[random_num_2]
            random_loc_list = []
            for j in range(num_point):
                while True:
                    random_loc = random.randint(0, length)
                    if random_loc not in random_loc_list:
                        random_loc_list.append(random_loc)
                        break
                medium_gene = child_individual_1[random_loc]
                child_individual_1[random_loc] = child_individual_2[random_loc]
                child_individual_2[random_loc] = medium_gene
            child_matrix.append(child_individual_1)
            child_matrix.append(child_individual_2)
        child_matrix = np.array(child_matrix)
        child_matrix = np.concatenate([parent_matrix, child_matrix])
        return child_matrix
    
    
    def __uniform_crossover(self, length, parent_matrix):
        '''
        均匀交叉
        
        
        Uniform Crossover
        '''
        num_of_parents = len(parent_matrix)
        child_matrix = []
        child_population = self.population - num_of_parents
        
        while len(child_matrix) < child_population:
            while True:
                random_num_1 = random.randint(0, num_of_parents)
                random_num_2 = random.randint(0, num_of_parents)
                if random_num_1 != random_num_2:
                    break
            child_individual_1 = parent_matrix[random_num_1]
            child_individual_2 = parent_matrix[random_num_2]
            for j in range(length):
                crossover_possibility = random.random()
                if crossover_possibility <= self.__crossover_rate:
                    medium_gene = child_individual_1[j]
                    child_individual_1[j] = child_individual_2[j]
                    child_individual_2[j] = medium_gene
            child_matrix.append(child_individual_1)
            child_matrix.append(child_individual_2)
        child_matrix = np.array(child_matrix)
        child_matrix = np.concatenate([parent_matrix, child_matrix])
        return child_matrix
    
    
    def __st(self, parent_matrix, fitness, num_dead):
        '''
        随机竞争选择
        num_dead: 死亡的数量
        
        
        Stochastic Tournament
        num_dead: the number of dead individual
        '''
        num_of_parents = len(parent_matrix)
        
        for i in range(num_dead):
            while True:
                random_num_1 = random.randint(0, num_of_parents - i)
                random_num_2 = random.randint(0, num_of_parents - i)
                if random_num_1 != random_num_2:
                    break
            if fitness(parent_matrix[random_num_1]) <= fitness(parent_matrix[random_num_2]):
                parent_matrix = np.delete(parent_matrix, random_num_1, axis=0)
            else:
                parent_matrix = np.delete(parent_matrix, random_num_2, axis=0)
        return parent_matrix
    
    
    def __rw(self, parent_matrix, fitness, num_alive):
        '''
        轮盘赌选择
        
        
        Roulette Wheel Selection
        '''
        fitness_list = []
        child_matrix = []
        
        for parent_individual in parent_matrix:
            fitness_list.append(fitness(parent_individual))
        max_fitness = max(fitness_list)
        parent_matrix = list(parent_matrix)
        
        while len(child_matrix) < num_alive:
            ind = np.random.randint(0, len(parent_matrix))
            if np.random.rand() > fitness_list[ind] / max_fitness or parent_matrix[ind] in child_matrix:
                pass
            else:
                child_matrix.append(parent_matrix[ind])
        parent_matrix = np.array(child_matrix)
        return parent_matrix
    
    
    def fit(self, length, fitness):
        '''
        进行遗传算法模拟
        
        参数
        ----
        length：染色体长度
        fitness：适应度函数
        
        
        simulate
        
        Parameters
        ----------
        length: the length of chromosome
        fitness: fitness function
        '''
        self.__fitness = fitness
        if self.random_type == 'random':
            parent_matrix = np.random.rand(self.population, length)
        elif self.random_type == 'randint':
            parent_matrix = np.random.randint(self.__begin, self.__end, size=(self.population, length))
        else:
            raise Exception("'random_type' must be one of 'random' and 'randint'.")
        
        num_alive = self.population * self.__select_rate
        num_dead = int(self.population - num_alive)
        num_point = int(length * self.__crossover_rate)
        
        for i in range(self.iteration - 1):
            if self.select == 'rw':
                parent_matrix = GA.__rw(self, parent_matrix=parent_matrix, fitness=fitness, num_alive=num_alive)
            elif self.select == 'st':
                parent_matrix = GA.__st(self, parent_matrix=parent_matrix, fitness=fitness, num_dead=num_dead)
            elif type(self.select).__name__ == 'function':
                parent_matrix == self.select(parent_matrix, fitness)
            
            if self.crossover == 'uniform':
                parent_matrix = GA.__uniform_crossover(self, length=length, parent_matrix=parent_matrix)
            elif self.crossover == 'point':
                parent_matrix = GA.__point_crossover(self, num_point=num_point, length=length, parent_matrix=parent_matrix)
            elif type(self.select).__name__ == 'function':
                parent_matrix == self.select(parent_matrix, fitness)
        
        if self.select == 'rw':
            GA.__rw(self, parent_matrix=parent_matrix, fitness=fitness, num_alive=num_alive)
        elif self.select == 'st':
            GA.__st(self, parent_matrix=parent_matrix, fitness=fitness, num_dead=num_dead)
        elif type(self.select).__name__ == 'function':
            parent_matrix == self.select(parent_matrix, fitness)
        
        self.dom = parent_matrix
        
        dominance_fitness = []
        for individual in parent_matrix:
            dominance_fitness.append(fitness(individual))
        self.dom_fitness = np.array(dominance_fitness)



class SA():
    '''
    模拟退火算法
    默认以评估函数select的最小值为最优值
    
    参数
    ----
    anneal：浮点数类型或函数类型，可选，退火方法，若为浮点数，则按T = anneal * T退火，默认为0.9
    step：可选，步长倍率，每次生成的步长为k乘一个属于(-1, 1)的随机数，默认为1
    n：整型，可选，等温时迭代次数，默认为10
    random_state：整型，可选，随机种子
    
    属性
    ----
    anneal：退火方法
    step：步长倍率
    n：等温时迭代次数
    solution：最优解
    value：最优解的函数值
    
    
    Simulated Annealing Algorithm
    By default, the minimum value of the evaluation function 'select' is the optimal value
    
    Parameters
    ----------
    anneal: float or function, callable, annealing method, if type is float, it will be annealed with T = anneal * T, default=0.9
    step: callable, step, each generated step = k * a random number belonging to (-1, 1), default=1
    n: int, callable, isothermal iterations, default=10
    random_state: int, callable, random seed
    
    Attributes
    ----------
    anneal: annealing method
    step: step
    n: isothermal iterations
    solution: optimal solution
    value: function value of optimal solution
    '''
    def __init__(self, anneal=0.9, step=1, n=10, random_state=None):
        self.anneal = anneal
        self.step = step
        self.n = n
        np.random.seed(random_state)
    
    
    def fit(self, initial, T, T0, select):
        '''
        参数
        ----
        initial：列表，初始解
        T：初始温度
        T0：退火温度
        select：函数，评估函数
        
        
        Parameters
        ----------
        initial: list, initial solution
        T: initial temperature
        T0: annealing temperature
        select: function, evaluation function
        '''
        initial = np.array(initial, dtype=np.float)
        dim = len(initial)
        while T > T0:
            for i in range(self.n):
                random_x = 2 * self.step * (np.random.rand(dim) - 0.5)
                initial_copy = initial.copy()
                initial_copy += random_x
                if select(initial_copy) < select(initial):
                    initial = initial_copy
                else:
                    pro = random.random()
                    if pro < np.e**(-(select(initial_copy) - select(initial)) / T):
                        initial = initial_copy
            
            if type(self.anneal).__name__ == 'float':
                T *= self.anneal
            elif type(self.anneal).__name__ == 'function':
                T = self.anneal(T)
            else:
                raise Exception("Type of 'anneal' must be one of 'float' and 'function'.")
        
        self.solution = initial
        self.value = select(initial)


'''
本模块储存着常用算符
调用模块中的算符时应注意：
1、算符作用的多元函数的自变量需为列表(list、tuple、ndarray均可，下同)
2、算符作用的标量函数的输出值需为单个值
3、算符作用的矢量函数的输出值需为列表

This module stores the common operators
When calling operators in the module, pay attention to:
1. The argument of the multivariate function acted by the operators should be list (list, tuple, ndarray, the same below)
2. The output value of scalar function acted by the operators should be a single value
3. The output value of vector function acted by the operators should be list
'''

import numpy as np
from scipy import constants as C


class Dif():
    '''
    对多元函数的求导算子，对矢量函数则相当于对每个分量函数都进行了相同的求导
    多元矢量函数可以分解为多个标量函数
    无法将二维列表识别成一行作为一个自变量，需要循环
    
    参数
    ----
    dim：整型，可选，对第dim个自变量求导，默认为0
    acc：浮点数类型，可选，求导的精度，默认为0.05
    
    
    derivative operator for multivariate function( same derivation to each component)
    Multivariate vector functions can be decomposed into several scalar functions
    Unable to recognize the two-dimensional list as an independent variable, loop is required
    
    Parameters
    ----------
    dim: int, callable, derivation of the dim independent variable, default=0
    acc: float, callable, accuracy of derivation, default=0.05
    '''
    def __init__(self, dim=0, acc=0.05):
        self.dim = dim
        self.acc = acc
    

    def fit(self, f):
        '''
        参数
        ----
        f:函数
        函数的自变量x是列表

        返回值：一个导函数
        
        
        Parameter
        ---------
        f: function
        argument x of function should be a list
        
        return: a derivative function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            x[self.dim] += self.acc
            func1 = f(x)
            if type(func1).__name__ == 'tuple' or 'list':
                func1 = np.array(func1)
            x[self.dim] -= 2 * self.acc
            func2 = f(x)
            if type(func2).__name__ == 'tuple' or 'list':
                func2 = np.array(func2)
            de = (func1 - func2) / (self.acc + self.acc)
            return de
        return obj



class Del():
    '''
    ▽算子
    被作用函数的自变量x是列表
    
    参数
    ----
    acc：浮点数类型，可选，求导的精度，默认为0.05
    
    
    nabla operator ▽
    the argument x of the affected function should be a list
    
    Parameters
    ----------
    acc: float, callable, accuracy of derivation, default=0.05
    '''
    def __init__(self, acc=0.05):
        self.acc = acc
    

    def grad(self, f, dim=None):
        '''
        求标量函数梯度：▽f(r)
        
        参数
        ----
        f：函数，要求函数f返回一个数值
        dim：整型，求导的维度，默认全部求导
        
        返回值：一个梯度函数
        
        
        gradient of scalar function: ▽f(r)
        
        Parameters
        ----------
        f: function, the function should return a number
        dim: int, dimensions for derivation, default: all
        
        return: a gradient function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            if not dim:
                dim = len(x)
            result = []
            for i in range(dim):
                x[i] += self.acc
                func1 = f(x)
                x[i] -= 2 * self.acc
                func2 = f(x)
                x[i] += self.acc
                de = (func1 - func2) / (self.acc + self.acc)
                result.append(de)
            result = np.array(result)
            return result
        return obj
    

    def dot(self, f):
        '''
        ▽点乘矢量函数：▽·f(r)
        
        参数
        ----
        f：函数，要求函数f返回一个列表

        返回值：一个新函数
        
        
        dot product between ▽ and vector function: ▽·f(r)
        
        Parameter
        ---------
        f: function, the function should return a list
        
        return: a new function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            result = []
            for i in range(len(x)):
                x[i] += self.acc
                func1 = f(x)[i]
                x[i] -= 2*self.acc
                func2 = f(x)[i]
                x[i] += self.acc
                de = (func1 - func2) / (self.acc + self.acc)
                result.append(de)
            result = np.array(result)
            return result
        return obj
    

    def __diff2forc(self, a, b, f, x):
        '''
        一个中介函数，方便后文程序使用
        '''
        x[b] += self.acc
        func1 = f(x)[a]
        x[b] -= 2*self.acc
        func2 = f(x)[a]
        x[b] += self.acc
        de = (func1 - func2) / (self.acc + self.acc)
        return de
    
    
    def cross(self, f):
        '''
        ▽叉乘矢量函数：▽×f(r)
        
        参数
        ----
        f：函数，要求函数是三维矢量函数

        返回值：一个新函数
        
        
        ▽ cross vector function: ▽×f(r)
        
        Parameter
        ---------
        f: function, the function should be a three-dimension vector function
        
        return: a new function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            result = np.array([
                Del.__diff2forc(self, 2, 1, f, x) -\
                Del.__diff2forc(self, 1, 2, f, x),
                Del.__diff2forc(self, 0, 2, f, x) -\
                Del.__diff2forc(self, 2, 0, f, x),
                Del.__diff2forc(self, 1, 0, f, x) -\
                Del.__diff2forc(self, 0, 1, f, x)])
            return result
        return obj



class Laplace():
    '''
    △算子 = ▽**2
    被作用函数的自变量x是列表
    
    参数
    ----
    acc：浮点数类型，可选，求导的精度，默认为0.05
    
    
    Laplace operator: △ = ▽**2
    the argument x of the affected function should be a list
    
    Parameter
    ---------
    acc: float, callable, accuracy of derivation, default=0.05
    '''
    def __init__(self, acc=0.05):
        self.acc = acc
    

    def fit(self, f):
        '''
        参数
        ----
        f：函数

        返回值：一个新函数
        
        
        Parameter
        ---------
        f: function
        
        return: a new function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            result = 0
            for i in range(len(x)):
                func1 = 2 * f(x)
                x[i] += 2 * self.acc
                func2 = f(x)
                x[i] -= 4 * self.acc
                func3 = f(x)
                x[i] += 2 * self.acc
                de = (func2 + func3 - 2 * func1) / 4 / self.acc**2
                result += de
            return result
        return obj



class H():
    '''
    哈密顿算符：H = - hr**2 / 2m * ▽**2 + U
    
    参数
    ----
    m：数，例子质量
    U：数或函数，势能
    acc：浮点数类型，可选，求导的精度，默认为0.05
    
    
    Hamilton: H = - hr**2 / 2m * ▽**2 + U
    
    Parameters
    ----------
    m: number, the mass of the particle
    U: number or function, potential energy
    acc: float, callable, accuracy of derivation, default=0.05
    '''
    def __init__(self, m, U, acc=0.05):
        self.m = m
        self.U = U
        self.acc = acc
    
    
    def fit(self, f):
        '''
        参数
        ----
        f：函数
        
        
        Parameter
        ---------
        f: function
        '''
        def obj(x):
            x = np.array(x, dtype=float)
            result = 0
            for i in range(len(x)):
                func1 = 2 * f(x)
                x[i] += 2 * self.acc
                func2 = f(x)
                x[i] -= 4 * self.acc
                func3 = f(x)
                x[i] += 2 * self.acc
                de = (func2 + func3 - 2 * func1) / 4 / self.acc**2
                result += de
            result *= (C.h / (2 * np.pi))**2 / (2 * self.m)
            
            if type(self.U).__name__ == 'function':
                result += self.U(x) * f(x)
            else:
                result += self.U * f(x)
            return result